import unittest
from datetime import datetime, timedelta

import pytest

from icalendar.prop import vPeriod


class TestProp(unittest.TestCase):
    def test_one_day(self):
        # One day in exact datetimes
        per = (datetime(2000, 1, 1), datetime(2000, 1, 2))
        assert vPeriod(per).to_ical() == b"20000101T000000/20000102T000000"

        per = (datetime(2000, 1, 1), timedelta(days=31))
        assert vPeriod(per).to_ical() == b"20000101T000000/P31D"

    def test_roundtrip(self):
        p = vPeriod.from_ical("20000101T000000/20000102T000000")
        assert p == (datetime(2000, 1, 1, 0, 0), datetime(2000, 1, 2, 0, 0))
        assert vPeriod(p).to_ical() == b"20000101T000000/20000102T000000"

        assert vPeriod.from_ical("20000101T000000/P31D") == (
            datetime(2000, 1, 1, 0, 0),
            timedelta(31),
        )

    def test_round_trip_with_absolute_time(self):
        p = vPeriod.from_ical("20000101T000000Z/20000102T000000Z")
        assert vPeriod(p).to_ical() == b"20000101T000000Z/20000102T000000Z"

    def test_bad_input(self):
        self.assertRaises(ValueError, vPeriod.from_ical, "20000101T000000/Psd31D")


def test_timezoned(tzp):
    start = tzp.localize(datetime(2000, 1, 1), "Europe/Copenhagen")
    end = tzp.localize(datetime(2000, 1, 2), "Europe/Copenhagen")
    per = (start, end)
    assert vPeriod(per).to_ical() == b"20000101T000000/20000102T000000"
    assert vPeriod(per).params["TZID"] == "Europe/Copenhagen"


def test_timezoned_with_timedelta(tzp):
    p = vPeriod(
        (tzp.localize(datetime(2000, 1, 1), "Europe/Copenhagen"), timedelta(days=31))
    )
    assert p.to_ical() == b"20000101T000000/P31D"


@pytest.mark.parametrize(
    ("period", "error"),
    [
        (("20000101T000000", datetime(2000, 1, 2)), TypeError),
        ((datetime(2000, 1, 1), "20000102T000000"), TypeError),
        ((datetime(2000, 1, 2), datetime(2000, 1, 1)), ValueError),
        ((datetime(2000, 1, 2), timedelta(-1)), ValueError),
    ],
)
def test_invalid_parameters(period, error):
    """The parameters are of wrong type or of wrong order."""
    with pytest.raises(error):
        vPeriod(period)



"""We can get time information that is actually of the WRONG value.

We can still parse it and correct the value parameter.
"""
from datetime import date, datetime

import pytest

from icalendar import Event
from icalendar.compatibility import ZoneInfo


@pytest.mark.parametrize(
    ("string", "expected_params", "expected_property_value"),
    [
        (":20150217", {}, date(2015, 2, 17)),
        (";VALUE=DATETIME:20150217", {"VALUE": "DATETIME"}, date(2015, 2, 17)),
        (";VALUE=DATE:20150217T095800", {"VALUE": "DATE"}, datetime(2015, 2, 17, 9, 58, 0)),
        (";VALUE=DATE:20150217T095800Z", {"VALUE": "DATE"}, datetime(2015, 2, 17, 9, 58, 0, tzinfo=ZoneInfo("UTC"))),
        (";TZID=Europe/Berlin:20150217", { "TZID": "Europe/Berlin"}, datetime(2015, 2, 17, tzinfo=ZoneInfo("Europe/Berlin"))),
        (";TZID=Europe/Berlin:20150217T095800Z", {"TZID": "Europe/Berlin"}, datetime(2015, 2, 17, 9, 58, 0, tzinfo=ZoneInfo("Europe/Berlin"))),
    ]
)
def test_wrong_dates_are_converted(string, expected_params, expected_property_value):
    """We convert these dates to the correct value if they turn up.

    We want to make sure that these are working, have values and do not throw exceptions.
    Information should not get lost except if it is contradictory.
    So, the tests can be changed if a different resolution is needed.

    The problem: The VALUE parameter might not correspond to the actual value.
    However, we correctly parse the value and do not lose information.
    """
    event = Event.from_ical(f"""
BEGIN:VEVENT
DTSTART{string}
END:VEVENT
""")
    dt = event.get("dtstart")
    test = f"{dt} should be {expected_property_value} with {expected_params}"
    assert dt.params == expected_params, test
    assert dt.dt == expected_property_value, test



from datetime import datetime

import pytest

from icalendar.prop import vDatetime


def test_to_ical():
    assert vDatetime(datetime(2001, 1, 1, 12, 30, 0)).to_ical() == b"20010101T123000"


def test_from_ical():
    assert vDatetime.from_ical("20000101T120000") == datetime(2000, 1, 1, 12, 0)
    assert vDatetime.from_ical("20010101T000000") == datetime(2001, 1, 1, 0, 0)


def test_to_ical_utc(tzp):
    dutc = tzp.localize_utc(datetime(2001, 1, 1, 12, 30, 0))
    assert vDatetime(dutc).to_ical() == b"20010101T123000Z"


def test_to_ical_utc_1899(tzp):
    dutc = tzp.localize_utc(datetime(1899, 1, 1, 12, 30, 0))
    assert vDatetime(dutc).to_ical() == b"18990101T123000Z"


def test_bad_ical():
    with pytest.raises(ValueError):
        vDatetime.from_ical("20010101T000000A")


def test_roundtrip():
    utc = vDatetime.from_ical("20010101T000000Z")
    assert vDatetime(utc).to_ical() == b"20010101T000000Z"


def test_transition(tzp):
    # 1 minute before transition to DST
    dat = vDatetime.from_ical("20120311T015959", "America/Denver")
    assert dat.strftime("%Y%m%d%H%M%S %z") == "20120311015959 -0700"

    # After transition to DST
    dat = vDatetime.from_ical("20120311T030000", "America/Denver")
    assert dat.strftime("%Y%m%d%H%M%S %z") == "20120311030000 -0600"

    dat = vDatetime.from_ical("20101010T000000", "Europe/Vienna")
    assert vDatetime(dat).to_ical() == b"20101010T000000"



"""Test the alarm time computation.

Events can have alarms.
Alarms can be in this state:

- active - the user wants the alarm to pop up
- acknowledged - the user no longer wants the alarm
- snoozed - the user moved that alarm to another time

The alarms can only work on the properties of the event like
DTSTART, DTEND, and DURATION.

See https://github.com/collective/icalendar/issues/716
"""

from datetime import date, datetime, timedelta, timezone

import pytest

from icalendar.alarms import Alarms
from icalendar.cal.alarm import Alarm
from icalendar.cal.event import Event
from icalendar.error import IncompleteAlarmInformation
from icalendar.prop import vDatetime
from icalendar.tools import normalize_pytz

UTC = timezone.utc
EXAMPLE_TRIGGER = datetime(1997, 3, 17, 13, 30, tzinfo=UTC)


def test_absolute_alarm_time_rfc_example(alarms):
    """Check that the absolute alarm is recognized.

    The following example is for a "VALARM" calendar component
    that specifies an audio alarm that will sound at a precise time
    and repeat 4 more times at 15-minute intervals:
    """
    a = Alarms(alarms.rfc_5545_absolute_alarm_example)
    times = a.times
    assert len(times) == 5
    for i, t in enumerate(times):
        assert t.trigger == EXAMPLE_TRIGGER + timedelta(minutes=15 * i)


alarm_1 = Alarm()
alarm_1.add("TRIGGER", EXAMPLE_TRIGGER)
alarm_2 = Alarm()
alarm_2["TRIGGER"] = vDatetime(EXAMPLE_TRIGGER)


@pytest.mark.parametrize("alarm", [alarm_1, alarm_2])
def test_absolute_alarm_time_with_vDatetime(alarm):
    """Check that the absolute alarm is recognized.

    The following example is for a "VALARM" calendar component
    that specifies an audio alarm that will sound at a precise time
    and repeat 4 more times at 15-minute intervals:
    """
    a = Alarms(alarm)
    times = a.times
    assert len(times) == 1
    assert times[0].trigger == EXAMPLE_TRIGGER


alarm_incomplete_1 = Alarm()
alarm_incomplete_1.TRIGGER = timedelta(hours=2)
alarm_incomplete_1.DURATION = timedelta(hours=1)
alarm_incomplete_2 = Alarm()
alarm_incomplete_2.TRIGGER = timedelta(hours=2)
alarm_incomplete_2.REPEAT = 100


@pytest.mark.parametrize("alarm", [alarm_incomplete_1, alarm_incomplete_2])
def test_alarm_has_only_one_of_repeat_or_duration(alarm):
    """This is an edge case and we should ignore the repetition."""
    a = Alarms(alarm)
    a.set_start(datetime(2027, 12, 2))
    assert len(a.times) == 1


@pytest.fixture(params=[(0, timedelta(minutes=-30)), (1, timedelta(minutes=-25))])
def alarm_before_start(calendars, request):
    """An example alarm relative to the start of a component."""
    index, td = request.param
    alarm = calendars.alarm_etar_future.subcomponents[-1].subcomponents[index]
    assert isinstance(alarm, Alarm)
    assert alarm.get("TRIGGER").dt == td
    alarm.test_td = td
    return alarm


def test_cannot_compute_relative_alarm_without_start(alarm_before_start):
    """We have an alarm without a start of a component."""
    with pytest.raises(IncompleteAlarmInformation) as e:
        Alarms(alarm_before_start).times  # noqa: B018
    assert (
        e.value.args[0]
        == f"Use {Alarms.__name__}.{Alarms.set_start.__name__} because at least one alarm is relative to the start of a component."
    )


@pytest.mark.parametrize(
    ("dtstart", "timezone", "trigger"),
    [
        (
            datetime(2024, 10, 29, 13, 10),
            "UTC",
            datetime(2024, 10, 29, 13, 10, tzinfo=UTC),
        ),
        (date(2024, 11, 16), None, datetime(2024, 11, 16, 0, 0)),
        (
            datetime(2024, 10, 29, 13, 10),
            "Asia/Singapore",
            datetime(2024, 10, 29, 5, 10, tzinfo=UTC),
        ),
        (datetime(2024, 10, 29, 13, 20), None, datetime(2024, 10, 29, 13, 20)),
    ],
)
def test_can_complete_relative_calculation_if_a_start_is_given(
    alarm_before_start, dtstart, timezone, trigger, tzp
):
    """The start is given and required."""
    start = dtstart if timezone is None else tzp.localize(dtstart, timezone)
    alarms = Alarms(alarm_before_start)
    alarms.set_start(start)
    assert len(alarms.times) == 1
    time = alarms.times[0]
    expected_trigger = normalize_pytz(trigger + alarm_before_start.test_td)
    assert time.trigger == expected_trigger


@pytest.mark.parametrize("dtstart", [date(1998, 10, 1), date(2023, 12, 31)])
def test_start_as_date_with_delta_as_date_stays_date(alarms, dtstart):
    """If we have an alarm with a day delta and the event is a day event, we should stay as a date."""
    a = Alarms(alarms.start_date)
    a.set_start(dtstart)
    assert len(a.times) == 1
    assert a.times[0].trigger == dtstart - timedelta(days=2)


def test_cannot_compute_relative_alarm_without_end(alarms):
    """We have an alarm without an end of a component."""
    with pytest.raises(IncompleteAlarmInformation) as e:
        Alarms(alarms.rfc_5545_end).times  # noqa: B018
    assert (
        e.value.args[0]
        == f"Use {Alarms.__name__}.{Alarms.set_end.__name__} because at least one alarm is relative to the end of a component."
    )


@pytest.mark.parametrize(
    ("dtend", "timezone", "trigger"),
    [
        (
            datetime(2024, 10, 29, 13, 10),
            "UTC",
            datetime(2024, 10, 29, 13, 10, tzinfo=UTC),
        ),
        (date(2024, 11, 16), None, date(2024, 11, 16)),
        (
            datetime(2024, 10, 29, 13, 10),
            "Asia/Singapore",
            datetime(2024, 10, 29, 5, 10, tzinfo=UTC),
        ),
        (datetime(2024, 10, 29, 13, 20), None, datetime(2024, 10, 29, 13, 20)),
    ],
)
def test_can_complete_relative_calculation(alarms, dtend, timezone, trigger, tzp):
    """The start is given and required."""
    start = dtend if timezone is None else tzp.localize(dtend, timezone)
    alarms = Alarms(alarms.rfc_5545_end)
    alarms.set_end(start)
    assert len(alarms.times) == 1
    time = alarms.times[0]
    expected_trigger = normalize_pytz(trigger - timedelta(days=2))
    assert time.trigger == expected_trigger


@pytest.mark.parametrize("dtend", [date(1998, 10, 1), date(2023, 12, 31)])
def test_end_as_date_with_delta_as_date_stays_date(alarms, dtend):
    """If we have an alarm with a day delta and the event is a day event, we should stay as a date."""
    a = Alarms(alarms.rfc_5545_end)
    a.set_end(dtend)
    assert len(a.times) == 1
    assert a.times[0].trigger == dtend - timedelta(days=2)


def test_add_multiple_alarms(alarms):
    """We can add multiple alarms."""
    a = Alarms()
    a.add_alarm(alarms.start_date)
    a.add_alarm(alarms.rfc_5545_end)
    a.add_alarm(alarms.rfc_5545_absolute_alarm_example)
    with pytest.raises(IncompleteAlarmInformation):
        a.times  # noqa: B018
    a.set_start(datetime(2012, 3, 5))
    with pytest.raises(IncompleteAlarmInformation):
        a.times  # noqa: B018
    a.set_end(datetime(2012, 3, 5))
    assert len(a.times) == 7


def test_alarms_from_event_have_right_times(calendars):
    """We can collect from an event."""
    event = calendars.alarm_etar_future.subcomponents[-1]
    a = Alarms(event)
    assert len(a.times) == 3
    assert a.times[0].parent == event


def test_cannot_set_the_event_twice(calendars):
    """We cannot set an event twice. This make the state ambiguous."""
    event = calendars.alarm_etar_future.subcomponents[-1]
    a = Alarms()
    a.add_component(event)
    a.add_component(event)  # same component is ok
    with pytest.raises(ValueError):
        a.add_component(calendars.alarm_google_future.subcomponents[-1])


@pytest.mark.parametrize(
    ("calendar", "index", "count", "message"),
    [
        ("alarm_etar_future", -1, 3, "Etar (1): we just created the alarm"),
        ("alarm_etar_notification", -1, 2, "Etar (2): the notification popped up"),
        (
            "alarm_etar_notification_clicked",
            -1,
            0,
            "Etar (3): the notification was dismissed",
        ),  # TODO: check that that is really true
        (
            "alarm_google_future",
            -1,
            4,
            "Google (1): we just created the event with alarms",
        ),
        (
            "alarm_google_acknowledged",
            -1,
            2,
            "Google (2): 2 alarms happened at the same time",
        ),
        ("alarm_thunderbird_future", -1, 2, "Thunderbird (1.1): 2 alarms are set"),
        (
            "alarm_thunderbird_snoozed_until_1457",
            -1,
            2,
            "Thunderbird (1.2): 2 alarms are snoozed to another time",
        ),
        (
            "alarm_thunderbird_closed",
            -1,
            0,
            "Thunderbird (1.3): all alarms are dismissed (closed)",
        ),
        ("alarm_thunderbird_2_future", -1, 2, "Thunderbird (2.1): 2 alarms active"),
        (
            "alarm_thunderbird_2_notification_popped_up",
            -1,
            2,
            "Thunderbird (2.2): one alarm popped up as a notification",
        ),
        (
            "alarm_thunderbird_2_notification_5_min_postponed",
            -1,
            2,
            "Thunderbird (2.3): 1 alarm active and one postponed by 5 minutes",
        ),
        (
            "alarm_thunderbird_2_notification_5_min_postponed_and_popped_up",
            -1,
            2,
            "Thunderbird (2.4): 1 alarm active and one postponed by 5 minutes and now popped up",
        ),
        (
            "alarm_thunderbird_2_notification_5_min_postponed_and_closed",
            -1,
            1,
            "Thunderbird (2.5): 1 alarm active and one postponed by 5 minutes and is now acknowledged",
        ),
    ],
)
def test_number_of_active_alarms_from_calendar_software(
    calendars, calendar, index, count, message
):
    """Check that we extract calculate the correct amount of active alarms."""
    event = calendars[calendar].subcomponents[index]
    a = Alarms(event)
    active_alarms = (
        a.active
    )  # We do not need to pass a timezone because the events have a timezone
    assert len(active_alarms) == count, (
        f"{message} - I expect {count} alarms active but got {len(active_alarms)}."
    )


three_alarms = Alarm()
three_alarms.REPEAT = 2
three_alarms.add("DURATION", timedelta(hours=1))  # 2 hours & 1 hour before
three_alarms.add("TRIGGER", -timedelta(hours=3))  # 3 hours before


@pytest.mark.parametrize(
    ("start", "acknowledged", "timezone", "count"),
    [
        (datetime(2024, 10, 10), datetime(2024, 10, 9), "UTC", 3),
        (datetime(2024, 10, 10, 12), datetime(2024, 10, 10, 9, 1), "UTC", 2),
        (datetime(2024, 10, 10, 12), datetime(2024, 10, 10, 10, 1), "UTC", 1),
        (datetime(2024, 10, 10, 12), datetime(2024, 10, 10, 11, 1), "UTC", 0),
        (
            datetime(2024, 10, 10, 12, tzinfo=timezone.utc),
            datetime(2024, 10, 10, 11, 1),
            None,
            0,
        ),
    ],
)
def test_number_of_active_alarms_with_moving_time(
    start, acknowledged, count, tzp, timezone
):
    """Check how many alarms are active after a time they are acknowledged."""
    a = Alarms()
    a.add_alarm(three_alarms)
    a.set_start(start)
    a.set_local_timezone(timezone)
    a.acknowledge_until(tzp.localize_utc(acknowledged))
    active = a.active
    assert len(active) == count


def test_incomplete_alarm_information_for_active_state(tzp):
    """Make sure we throw the right error."""
    a = Alarms()
    a.add_alarm(three_alarms)
    a.set_start(date(2017, 12, 1))
    a.acknowledge_until(tzp.localize_utc(datetime(2012, 10, 10, 12)))
    with pytest.raises(IncompleteAlarmInformation) as e:
        a.active  # noqa: B018
    assert (
        e.value.args[0]
        == f"A local timezone is required to check if the alarm is still active. Use Alarms.{Alarms.set_local_timezone.__name__}()."
    )


@pytest.mark.parametrize(
    "calendar_name",
    [
        "alarm_etar_future",
        "alarm_google_acknowledged",
        "alarm_thunderbird_closed",
        "alarm_thunderbird_future",
        "alarm_thunderbird_snoozed_until_1457",
    ],
)
def test_thunderbird_recognition(calendars, calendar_name):
    """Check if we correctly discover Thunderbird's alarm algorithm."""
    calendar = calendars[calendar_name]
    event = calendar.subcomponents[-1]
    assert isinstance(event, Event)
    assert event.is_thunderbird() == ("thunderbird" in calendar_name)


@pytest.mark.parametrize(
    "snooze",
    [
        datetime(2012, 10, 10, 11, 1),  # before everything
        datetime(2017, 12, 1, 10, 1),
        datetime(2017, 12, 1, 11, 1),
        datetime(2017, 12, 1, 12, 1),
        datetime(2017, 12, 1, 13, 1),  # snooze until after the start of the event
    ],
)
def test_snoozed_alarm_has_trigger_at_snooze_time(tzp, snooze):
    """When an alarm is snoozed, it pops up after the snooze time."""
    a = Alarms()
    a.add_alarm(three_alarms)
    a.set_start(datetime(2017, 12, 1, 13))
    a.set_local_timezone("UTC")
    snooze_utc = tzp.localize_utc(snooze)
    a.snooze_until(snooze_utc)
    active = a.active
    assert len(active) == 3
    for alarm in active:
        assert alarm.trigger >= snooze_utc


@pytest.mark.parametrize(
    ("event_index", "alarm_times"),
    [
        # Assume that we have the following event with an alarm set to trigger 15 minutes before the meeting:
        (1, ("20210302T101500",)),
        # When the alarm is triggered, the user decides to "snooze" it for 5 minutes.
        # The client acknowledges the original alarm and creates a new "snooze"
        # alarm as a sibling of, and relates it to, the original alarm (note that
        # both occurrences of "VALARM" reside within the same "parent" VEVENT):
        (2, ("20210302T102000",)),
        # When the "snooze" alarm is triggered, the user decides to "snooze" it
        # again for an additional 5 minutes. The client once again acknowledges
        # the original alarm, removes the triggered "snooze" alarm, and creates another
        # new "snooze" alarm as a sibling of, and relates it to, the original alarm
        # (note the different UID for the new "snooze" alarm):
        (3, ("20210302T102500",)),
        # When the second "snooze" alarm is triggered, the user decides to dismiss it.
        # The client acknowledges both the original alarm and the second "snooze" alarm:
        (4, ()),
    ],
)
def test_rfc_9074_alarm_times(events, event_index, alarm_times):
    """Test the examples from the RFC and their timing.

    Add times use America/New_York as timezone.
    """
    a = events[f"rfc_9074_example_{event_index}"].alarms
    assert len(a.active) == len(alarm_times)
    expected_alarm_times = {
        vDatetime.from_ical(t, "America/New_York") for t in alarm_times
    }
    computed_alarm_times = {alarm.trigger for alarm in a.active}
    assert expected_alarm_times == computed_alarm_times


def test_set_to_None():
    """acknowledge_until, snooze_until, set_local_timezone."""
    a = Alarms()
    a.set_start(None)
    a.set_end(None)
    a.set_local_timezone(None)
    a.acknowledge_until(None)
    a.snooze_until(None)
    assert vars(a) == vars(Alarms())


def test_delete_TRIGGER():
    """Delete the TRIGGER property."""
    a = Alarm()
    a.TRIGGER = datetime(2017, 12, 1, 10, 1)
    del a.TRIGGER
    assert a.TRIGGER is None



"""Issue #165 - Problem parsing a file with event recurring on weekdays

https://github.com/collective/icalendar/issues/165
"""


def test_issue_165_missing_event(calendars):
    events = list(calendars.issue_165_missing_event.walk("VEVENT"))
    assert len(events) == 1, "There was an event missing from the parsed events' list."



"""Test the property parameters.

See https://github.com/collective/icalendar/issues/798
"""

from datetime import datetime

import pytest

from icalendar import (
    CUTYPE,
    FBTYPE,
    PARTSTAT,
    RANGE,
    RELTYPE,
    vCalAddress,
    vDatetime,
    vPeriod,
    vText,
)
from icalendar.parser import Parameters
from icalendar.timezone.tzp import TZP


class Prop:
    params: Parameters

    def __init__(self, **parameters):
        """Create a new property."""
        self.params = Parameters(parameters)

    def to_ical(self) -> str:
        """Parameters to bytes to string."""
        return self.params.to_ical().decode("utf-8")

    from icalendar.param import (
        ALTREP,
        CN,
        CUTYPE,
        DELEGATED_FROM,
        DELEGATED_TO,
        RELTYPE,
    )


@pytest.fixture
def p():
    """Empty test property."""
    return Prop()


def test_set_altrep(p):
    p.ALTREP = "http://example.com"
    assert p.params == {"ALTREP": "http://example.com"}
    assert p.ALTREP == "http://example.com"


def test_altrep_must_be_quoted():
    """altrepparam = "ALTREP" "=" DQUOTE uri DQUOTE"""
    assert Prop(ALTREP="1234aA").to_ical() == 'ALTREP="1234aA"'


def test_del_altrep(p):
    """Del when empty"""
    del p.ALTREP
    assert p.params == {}
    assert p.ALTREP is None


def test_del_altrep_full(p):
    """Del when empty"""
    p.ALTREP = "http://example.com"
    del p.ALTREP
    assert p.params == {}
    assert p.ALTREP is None


def test_get_cutype(p):
    """The default is individual."""
    assert p.CUTYPE == "INDIVIDUAL"


def test_set_lowercase():
    p = Prop(CUTYPE="individual")
    assert p.CUTYPE == "INDIVIDUAL"
    p.CUTYPE = "unknown"
    assert p.CUTYPE == CUTYPE.UNKNOWN


def test_set_delegation_to_string(p):
    p.DELEGATED_FROM = "mailto:foo"
    assert p.DELEGATED_FROM == ("mailto:foo",)


def test_set_delegation_to_tuple(p):
    p.DELEGATED_TO = ("mailto:foo", "mailto:bar")
    assert p.DELEGATED_TO == ("mailto:foo", "mailto:bar")


def test_delete_delegation_to(p):
    p.DELEGATED_TO = ("mailto:foo", "mailto:bar")
    del p.DELEGATED_TO
    assert p.DELEGATED_TO == ()


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ((), ""),
        (("mailto:foo",), 'DELEGATED-TO="mailto:foo"'),
        (("mailto:foo", "mailto:bar"), 'DELEGATED-TO="mailto:foo","mailto:bar"'),
        (('mailto:"asd"',), "DELEGATED-TO=\"mailto:^'asd^'\""),
    ],
)
def test_serialize_delegation_to(p, value, expected):
    p.DELEGATED_TO = value
    assert p.to_ical() == expected


@pytest.mark.parametrize(
    ("index", "expected"),
    [
        (0, FBTYPE.BUSY_UNAVAILABLE),
        (1, FBTYPE.BUSY),
        (2, FBTYPE.FREE),
    ],
)
def test_get_fbtype(calendars, index, expected):
    fb = calendars.issue_798_freebusy.freebusy[index]
    p: vPeriod = fb["FREEBUSY"]
    if isinstance(p, list):
        p = p[0]
    assert expected == p.FBTYPE


@pytest.fixture
def addr():
    return vCalAddress("mailto:foo")


def test_partstat_get(addr: vCalAddress):
    """test the default partstat"""
    assert addr.PARTSTAT == "NEEDS-ACTION"


def test_set_the_partstat(addr: vCalAddress):
    addr.PARTSTAT = PARTSTAT.ACCEPTED
    assert addr.PARTSTAT == "ACCEPTED"


def test_this_and_future():
    assert vDatetime(datetime(2019, 12, 10)).RANGE is None


def test_this_and_future_set():
    d = vDatetime(datetime(2019, 12, 10))
    d.RANGE = RANGE.THISANDFUTURE
    assert d.params["RANGE"] == "THISANDFUTURE"


def test_rsvp_default(addr):
    assert not addr.RSVP


@pytest.mark.parametrize("rsvp", [True, False])
def test_set_rsvp(addr: vCalAddress, rsvp):
    addr.RSVP = rsvp
    assert rsvp == addr.RSVP
    assert addr.params["RSVP"] == ("TRUE" if rsvp else "FALSE")


def test_sent_by(addr: vCalAddress):
    assert addr.SENT_BY is None


def test_set_sent_by(addr: vCalAddress):
    addr.SENT_BY = "mailto:asd"
    assert addr.SENT_BY == "mailto:asd"
    assert addr.params["SENT-BY"] == "mailto:asd"
    assert addr.params.to_ical() == b'SENT-BY="mailto:asd"'


@pytest.mark.parametrize("tzid", [None, "Europe/Berlin"])
def test_tzid(tzid, tzp: TZP):
    """Check that the TZID parameter is determined by the datetime value."""
    dt = vDatetime(tzp.localize(datetime(2019, 12, 10), tzid))
    assert dt.TZID == tzid


@pytest.mark.parametrize(
    ("index", "reltype"),
    [
        (0, RELTYPE.PARENT),
        (1, RELTYPE.SIBLING),
    ],
)
def test_reltype_example(calendars, index, reltype):
    """The reltype parameter in the examples."""
    event = calendars.issue_798_related_to.events[index]
    print(event.to_ical().decode())
    r: vText = event["RELATED-TO"]
    print(r)
    print(r.params)
    assert reltype == r.RELTYPE


def test_set_reltype(p):
    p.RELTYPE = RELTYPE.CHILD
    assert p.RELTYPE == RELTYPE.CHILD
    assert p.params["RELTYPE"] == "CHILD"



"""Create a new method for the Calendar class.

See https://github.com/collective/icalendar/issues/596
"""

import pytest

from icalendar import Calendar, __version__


@pytest.mark.parametrize(
    ("name", "default_value"),
    [
        ("name", None),
        ("description", None),
        ("prodid", f"-//collective//icalendar//{__version__}//EN"),
        ("version", "2.0"),
        ("calscale", None),
        ("method", None),
        ("color", None),
    ],
)
def test_new_calendar_has_some_values(name, default_value):
    """When we create a new calendar, we want to have these values as default."""
    c = Calendar.new()
    assert c.get(name) == default_value
    assert default_value is None or name in c


@pytest.mark.parametrize(
    ("name", "value"),
    [
        ("name", "foo"),
        ("description", "bar"),
        ("prodid", "my-product"),
        ("version", "2.1"),
        ("calscale", "GREGORIAN"),
        ("method", "PUBLISH"),
        ("color", "#ff0000"),
    ],
)
def test_override_default_values(name, value):
    c = Calendar.new(**{name: value})
    assert c.get(name) == value


def test_default_getters():
    c = Calendar.new()
    assert c.calendar_name is None
    assert c.description is None
    assert c.version == "2.0"
    assert c.prodid == f"-//collective//icalendar//{__version__}//EN"
    assert c.calscale == "GREGORIAN"
    assert c.method == ""
    assert c.color == ""
    assert c.categories == []
    assert c.subcomponents == []


def test_categories():
    c = Calendar.new(categories=["foo", "bar"])
    assert c.categories == ["foo", "bar"]



""":rfc:`5545` VFREEBUSY component."""

from __future__ import annotations

import uuid
from datetime import date, datetime, timedelta
from typing import TYPE_CHECKING, Optional

from icalendar.attr import (
    CONCEPTS_TYPE_SETTER,
    LINKS_TYPE_SETTER,
    RELATED_TO_TYPE_SETTER,
    contacts_property,
    create_single_property,
    organizer_property,
    uid_property,
    url_property,
)
from icalendar.cal.component import Component

if TYPE_CHECKING:
    from icalendar.prop import vCalAddress


class FreeBusy(Component):
    """
        A "VFREEBUSY" calendar component is a grouping of component
        properties that represents either a request for free or busy time
        information, a reply to a request for free or busy time
        information, or a published set of busy time information.

    Examples:
        Create a new FreeBusy:

            >>> from icalendar import FreeBusy
            >>> free_busy = FreeBusy.new()
            >>> print(free_busy.to_ical())
            BEGIN:VFREEBUSY
            DTSTAMP:20250517T080612Z
            UID:d755cef5-2311-46ed-a0e1-6733c9e15c63
            END:VFREEBUSY

    """

    name = "VFREEBUSY"

    required = (
        "UID",
        "DTSTAMP",
    )
    singletons = (
        "CONTACT",
        "DTSTART",
        "DTEND",
        "DTSTAMP",
        "ORGANIZER",
        "UID",
        "URL",
    )
    multiple = (
        "ATTENDEE",
        "COMMENT",
        "FREEBUSY",
        "RSTATUS",
    )
    uid = uid_property
    url = url_property
    organizer = organizer_property
    contacts = contacts_property
    start = DTSTART = create_single_property(
        "DTSTART",
        "dt",
        (datetime, date),
        date,
        'The "DTSTART" property for a "VFREEBUSY" specifies the inclusive start of the component.',  # noqa: E501
    )
    end = DTEND = create_single_property(
        "DTEND",
        "dt",
        (datetime, date),
        date,
        'The "DTEND" property for a "VFREEBUSY" calendar component specifies the non-inclusive end of the component.',  # noqa: E501
    )

    @property
    def duration(self) -> Optional[timedelta]:
        """The duration computed from start and end."""
        if self.DTSTART is None or self.DTEND is None:
            return None
        return self.DTEND - self.DTSTART

    @classmethod
    def new(
        cls,
        /,
        comments: list[str] | str | None = None,
        concepts: CONCEPTS_TYPE_SETTER = None,
        contacts: list[str] | str | None = None,
        end: Optional[date | datetime] = None,
        links: LINKS_TYPE_SETTER = None,
        organizer: Optional[vCalAddress | str] = None,
        refids: list[str] | str | None = None,
        related_to: RELATED_TO_TYPE_SETTER = None,
        stamp: Optional[date] = None,
        start: Optional[date | datetime] = None,
        uid: Optional[str | uuid.UUID] = None,
        url: Optional[str] = None,
    ):
        """Create a new alarm with all required properties.

        This creates a new Alarm in accordance with :rfc:`5545`.

        Arguments:
            comments: The :attr:`~icalendar.Component.comments` of the component.
            concepts: The :attr:`~icalendar.Component.concepts` of the component.
            contacts: The :attr:`contacts` of the component.
            end: The :attr:`end` of the component.
            links: The :attr:`~icalendar.Component.links` of the component.
            organizer: The :attr:`organizer` of the component.
            refids: :attr:`~icalendar.Component.refids` of the component.
            related_to: :attr:`~icalendar.Component.related_to` of the component.
            stamp: The :attr:`DTSTAMP` of the component.
                If None, this is set to the current time.
            start: The :attr:`start` of the component.
            uid: The :attr:`uid` of the component.
                If None, this is set to a new :func:`uuid.uuid4`.
            url: The :attr:`url` of the component.

        Returns:
            :class:`FreeBusy`

        Raises:
            InvalidCalendar: If the content is not valid according to :rfc:`5545`.

        .. warning:: As time progresses, we will be stricter with the validation.
        """
        free_busy: FreeBusy = super().new(
            stamp=stamp if stamp is not None else cls._utc_now(),
            comments=comments,
            links=links,
            related_to=related_to,
            refids=refids,
            concepts=concepts,
        )
        free_busy.uid = uid if uid is not None else uuid.uuid4()
        free_busy.url = url
        free_busy.organizer = organizer
        free_busy.contacts = contacts
        free_busy.end = end
        free_busy.start = start

        if cls._validate_new:
            cls._validate_start_and_end(start, end)
        return free_busy


__all__ = ["FreeBusy"]



"""This tests the RDATE property."""

from __future__ import annotations

from datetime import date, datetime
from pprint import pprint
from typing import TYPE_CHECKING, Union

import pytest

from icalendar.cal.event import Event
from icalendar.cal.journal import Journal
from icalendar.cal.timezone import TimezoneDaylight, TimezoneStandard
from icalendar.cal.todo import Todo

if TYPE_CHECKING:
    from icalendar.cal.calendar import (
        Calendar,
    )

C_RDATE = Union[Event, Todo, Journal, TimezoneDaylight, TimezoneStandard]


@pytest.fixture(params=[Event, Todo, Journal, TimezoneDaylight, TimezoneStandard])
def c_rdate(request) -> C_RDATE:
    """Return a component that uses RDATE."""
    return request.param()


@pytest.fixture(
    params=[
        lambda _tzp: date(2019, 10, 11),
        lambda _tzp: datetime(2000, 1, 13, 12, 1),
        lambda tzp: tzp.localize_utc(datetime(2031, 12, 1, 23, 59)),
        lambda tzp: tzp.localize(datetime(1984, 1, 13, 13, 1), "Europe/Athens"),
        lambda _tzp: ((datetime(2000, 1, 13, 12, 1), datetime(2000, 1, 13, 12, 2))),
        lambda tzp: (
            (
                tzp.localize_utc(datetime(2001, 1, 13, 12, 1)),
                tzp.localize_utc(datetime(2001, 1, 13, 12, 2)),
            )
        ),
    ]
)
def rdate(request, tzp):
    """Possible values for an rdate."""
    return request.param(tzp)


def test_no_rdates_by_default(c_rdate):
    """We expect no rdate by default."""
    assert c_rdate.rdates == []


def test_set_and_retrieve_rdate(rdate, c_rdate):
    """Set the attribute and get the value."""
    c_rdate.add("RDATE", [rdate])
    result = [rdate if isinstance(rdate, tuple) else (rdate, None)]
    assert c_rdate.rdates == result


def test_get_example_0(calendars):
    """Test an example rdate."""
    cal: Calendar = calendars.rfc_5545_RDATE_example
    event = cal.events[0]
    assert event.rdates == [(datetime(1997, 7, 14, 12, 30), None)]


def test_get_example_1(calendars, tzp):
    """Test an example rdate."""
    cal: Calendar = calendars.rfc_5545_RDATE_example
    event = cal.events[1]
    assert event.rdates == [(tzp.localize_utc(datetime(1997, 7, 14, 12, 30)), None)]


def test_get_example_2(calendars, tzp):
    """Test an example rdate."""
    cal: Calendar = calendars.rfc_5545_RDATE_example
    event = cal.events[2]
    assert event.rdates == [
        (tzp.localize(datetime(1997, 7, 14, 8, 30), "America/New_York"), None)
    ]


def test_get_example_3(calendars, tzp):
    """Test an example rdate."""
    cal: Calendar = calendars.rfc_5545_RDATE_example
    event = cal.events[3]
    rdates_3 = [
        (
            tzp.localize_utc(datetime(1996, 4, 3, 2)),
            tzp.localize_utc(datetime(1996, 4, 3, 4)),
        ),
        (
            tzp.localize_utc(datetime(1996, 4, 4, 1)),
            tzp.localize_utc(datetime(1996, 4, 4, 4)),
        ),
    ]
    pprint(event.rdates)
    pprint(rdates_3)
    assert event.rdates == rdates_3


def d(i: int) -> tuple[date, None]:
    s = str(i)
    return (date(int(s[:4]), int(s[4:6].lstrip("0")), int(s[6:].lstrip("0"))), None)


RDATES_4 = list(
    map(
        d,
        (
            19970101,
            19970120,
            19970217,
            19970421,
            19970526,
            19970704,
            19970901,
            19971014,
            19971128,
            19971129,
            19971225,
        ),
    )
)


def test_get_example_4(calendars, tzp):
    """Test an example rdate."""
    cal: Calendar = calendars.rfc_5545_RDATE_example
    event = cal.events[4]
    pprint(event.rdates)
    pprint(RDATES_4)
    assert event.rdates == RDATES_4



"""This tests the properties of components and their types.

See https://github.com/collective/icalendar/issues/662
"""

from __future__ import annotations

from datetime import date, datetime, timedelta

import pytest

from icalendar import (
    Alarm,
    Event,
    IncompleteComponent,
    InvalidCalendar,
    Journal,
    Todo,
    vDatetime,
    vDDDTypes,
    vDuration,
)
from icalendar.compatibility import ZoneInfo


def prop(component: Event | Todo, prop: str) -> str:
    """Translate the end property.

    This allows us to run the same tests on Event and Todo.
    """
    if isinstance(component, Todo) and prop.upper() == "DTEND":
        return "DUE"
    return prop


@pytest.fixture(params=[Event, Todo])
def start_end_component(request):
    """The event to test."""
    return request.param()


@pytest.fixture(
    params=[
        datetime(2022, 7, 22, 12, 7),
        date(2022, 7, 22),
        datetime(2022, 7, 22, 13, 7, tzinfo=ZoneInfo("Europe/Paris")),
    ]
)
def dtstart(request, set_component_start, start_end_component):
    """Start of the event."""
    set_component_start(start_end_component, request.param)
    return request.param


def _set_component_start_init(component, start):
    """Create the event with the __init__ method."""
    d = dict(component)
    d["dtstart"] = vDDDTypes(start)
    component.clear()
    component.update(type(component)(d))


def _set_component_dtstart(component, start):
    """Create the event with the dtstart property."""
    component.DTSTART = start


def _set_component_start_attr(component, start):
    """Create the event with the dtstart property."""
    component.start = start


def _set_component_start_ics(component, start):
    """Create the event with the start property."""
    component.add("dtstart", start)
    ics = component.to_ical().decode()
    print(ics)
    component.clear()
    component.update(type(component).from_ical(ics))


@pytest.fixture(
    params=[
        _set_component_start_init,
        _set_component_start_ics,
        _set_component_dtstart,
        _set_component_start_attr,
    ]
)
def set_component_start(request):
    """Create a new event."""
    return request.param


def test_component_dtstart(dtstart, start_end_component):
    """Test the start of events."""
    assert dtstart == start_end_component.DTSTART


def test_event_start(dtstart, start_end_component):
    """Test the start of events."""
    assert start_end_component.start == dtstart


invalid_start_event_1 = Event()
invalid_start_event_1.add("dtstart", datetime(2022, 7, 22, 12, 7))
invalid_start_event_1.add("dtstart", datetime(2022, 7, 22, 12, 8))
invalid_start_event_2 = Event.from_ical(invalid_start_event_1.to_ical())
invalid_start_event_3 = Event()
invalid_start_event_3.add("DTSTART", (date(2018, 1, 1), date(2018, 2, 1)))
invalid_start_todo_1 = Todo(invalid_start_event_1)
invalid_start_todo_2 = Todo(invalid_start_event_2)
invalid_start_todo_3 = Todo(invalid_start_event_3)


@pytest.mark.parametrize(
    "invalid_event",
    [
        invalid_start_event_1,
        invalid_start_event_2,
        invalid_start_event_3,
        invalid_start_todo_1,
        invalid_start_todo_2,
        invalid_start_todo_3,
    ],
)
def test_multiple_dtstart(invalid_event):
    """Check that we get the right error."""
    with pytest.raises(InvalidCalendar):
        invalid_event.start  # noqa: B018
    with pytest.raises(InvalidCalendar):
        invalid_event.DTSTART  # noqa: B018


def test_no_dtstart(start_end_component):
    """DTSTART is optional.

    The following is REQUIRED if the component
    appears in an iCalendar object that doesn't
    specify the "METHOD" property; otherwise, it
    is OPTIONAL; in any case, it MUST NOT occur
    more than once.
    """
    assert start_end_component.DTSTART is None
    with pytest.raises(IncompleteComponent):
        start_end_component.start  # noqa: B018


@pytest.fixture(
    params=[
        datetime(2022, 7, 22, 12, 8),
        date(2022, 7, 23),
        datetime(2022, 7, 22, 14, 7, tzinfo=ZoneInfo("Europe/Paris")),
    ]
)
def dtend(request, set_component_end, start_end_component):
    """end of the event."""
    set_component_end(start_end_component, request.param)
    return request.param


def _set_component_end_init(component, end):
    """Create the event with the __init__ method."""
    d = dict(component)
    d[prop(component, "dtend")] = vDDDTypes(end)
    component.clear()
    component.update(type(component)(d))


def _set_component_end_property(component, end):
    """Create the event with the dtend property."""
    setattr(component, prop(component, "DTEND"), end)


def _set_component_end_attr(component, end):
    """Create the event with the dtend property."""
    component.end = end


def _set_component_end_ics(component, end):
    """Create the event with the end property."""
    component.add(prop(component, "DTEND"), end)
    ics = component.to_ical().decode()
    print(ics)
    component.clear()
    component.update(type(component).from_ical(ics))


@pytest.fixture(
    params=[
        _set_component_end_init,
        _set_component_end_ics,
        _set_component_end_property,
        _set_component_end_attr,
    ]
)
def set_component_end(request):
    """Create a new event."""
    return request.param


def test_component_end_property(dtend, start_end_component):
    """Test the end of events."""
    attr = prop(start_end_component, "DTEND")
    assert getattr(start_end_component, attr) == dtend


def test_component_end(dtend, start_end_component):
    """Test the end of events."""
    assert start_end_component.end == dtend


@pytest.mark.parametrize("attr", ["DTSTART", "DTEND"])
def test_delete_attr(start_end_component, dtstart, dtend, attr):
    attr = prop(start_end_component, attr)
    delattr(start_end_component, attr)
    assert getattr(start_end_component, attr) is None
    delattr(start_end_component, attr)


def _set_duration_vdddtypes(event: Event, duration: timedelta):
    """Set the vDDDTypes value"""
    event["DURATION"] = vDDDTypes(duration)


def _set_duration_add(event: Event, duration: timedelta):
    """Set the vDDDTypes value"""
    event.add("DURATION", duration)


def _set_duration_vduration(event: Event, duration: timedelta):
    """Set the vDDDTypes value"""
    event["DURATION"] = vDuration(duration)


@pytest.fixture(
    params=[_set_duration_vdddtypes, _set_duration_add, _set_duration_vduration]
)
def duration(start_end_component, dtstart, request):
    """... events have a DATE value type for the "DTSTART" property ...
    If such a "VEVENT" has a "DURATION"
    property, it MUST be specified as a "dur-day" or "dur-week" value.
    """
    duration = (
        timedelta(hours=1) if isinstance(dtstart, datetime) else timedelta(days=2)
    )
    request.param(start_end_component, duration)
    return duration


def test_start_and_duration(start_end_component, dtstart, duration):
    """Check calculation of end with duration."""
    dur = start_end_component.end - start_end_component.start
    assert dur == duration
    assert start_end_component.duration == duration


# The "VEVENT" is also the calendar component used to specify an
# anniversary or daily reminder within a calendar.  These events
# have a DATE value type for the "DTSTART" property instead of the
# default value type of DATE-TIME.  If such a "VEVENT" has a "DTEND"
# property, it MUST be specified as a DATE value also.
invalid_event_end_1 = Event()
invalid_event_end_1.add("DTSTART", datetime(2024, 1, 1, 10, 20))
invalid_event_end_1.add("DTEND", date(2024, 1, 1))
invalid_event_end_2 = Event()
invalid_event_end_2.add("DTEND", datetime(2024, 1, 1, 10, 20))
invalid_event_end_2.add("DTSTART", date(2024, 1, 1))
invalid_event_end_3 = Event()
invalid_event_end_3.add("DTEND", datetime(2024, 1, 1, 10, 20))
invalid_event_end_3.add("DTSTART", datetime(2024, 1, 1, 10, 20))
invalid_event_end_3.add("DURATION", timedelta(days=1))
invalid_event_end_4 = Event()
invalid_event_end_4.add("DTSTART", date(2024, 1, 1))
invalid_event_end_4.add("DURATION", timedelta(hours=1))

invalid_todo_end_1 = Todo()
invalid_todo_end_1.add("DTSTART", datetime(2024, 1, 1, 10, 20))
invalid_todo_end_1.add("DUE", date(2024, 1, 1))
invalid_todo_end_2 = Todo()
invalid_todo_end_2.add("DUE", datetime(2024, 1, 1, 10, 20))
invalid_todo_end_2.add("DTSTART", date(2024, 1, 1))
invalid_todo_end_3 = Todo()
invalid_todo_end_3.add("DUE", datetime(2024, 1, 1, 10, 20))
invalid_todo_end_3.add("DTSTART", datetime(2024, 1, 1, 10, 20))
invalid_todo_end_3.add("DURATION", timedelta(days=1))
invalid_todo_end_4 = Todo()
invalid_todo_end_4.add("DTSTART", date(2024, 1, 1))
invalid_todo_end_4.add("DURATION", timedelta(hours=1))


@pytest.mark.parametrize(
    ("invalid_component", "message"),
    [
        (
            invalid_event_end_1,
            "DTSTART and DTEND must be of the same type, either date or datetime.",
        ),
        (
            invalid_event_end_2,
            "DTSTART and DTEND must be of the same type, either date or datetime.",
        ),
        (
            invalid_event_end_3,
            "Only one of DTEND and DURATION may be in a VEVENT, not both.",
        ),
        (
            invalid_event_end_4,
            "When DTSTART is a date, DURATION must be of days or weeks.",
        ),
        (
            invalid_todo_end_1,
            "DTSTART and DUE must be of the same type, either date or datetime.",
        ),
        (
            invalid_todo_end_2,
            "DTSTART and DUE must be of the same type, either date or datetime.",
        ),
        (
            invalid_todo_end_3,
            "Only one of DUE and DURATION may be in a VTODO, not both.",
        ),
        (
            invalid_todo_end_4,
            "When DTSTART is a date, DURATION must be of days or weeks.",
        ),
    ],
)
@pytest.mark.parametrize("attr", ["start", "end"])
def test_invalid_event(invalid_component, message, attr):
    """Test that the end and start throuw the right error."""
    with pytest.raises(InvalidCalendar) as e:
        getattr(invalid_component, attr)
    assert e.value.args[0] == message


def test_event_duration_zero():
    """
    For cases where a "VEVENT" calendar component
    specifies a "DTSTART" property with a DATE-TIME value type but no
    "DTEND" property, the event ends on the same calendar date and
    time of day specified by the "DTSTART" property.
    """
    event = Event()
    event.start = datetime(2024, 10, 11, 10, 20)
    assert event.end == event.start
    assert event.duration == timedelta(days=0)


def test_event_duration_one_day():
    """
    For cases where a "VEVENT" calendar component
    specifies a "DTSTART" property with a DATE value type but no
    "DTEND" nor "DURATION" property, the event's duration is taken to
    be one day
    """
    event = Event()
    event.start = date(2024, 10, 11)
    assert event.end == event.start + timedelta(days=1)
    assert event.duration == timedelta(days=1)


def test_todo_duration_zero():
    """We do not know about the duration of a todo really."""
    todo = Todo()
    todo.start = datetime(2024, 10, 11, 10, 20)
    assert todo.end == todo.start
    assert todo.duration == timedelta(days=0)


def test_todo_duration_one_day():
    """The end is at the end of the day, excluding midnight.

    RFC 5545:
    The following is an example of a "VTODO" calendar
    component that needs to be completed before May 1st, 2007.  On
    midnight May 1st, 2007 this to-do would be considered overdue.
    """
    event = Event()
    event.start = date(2024, 10, 11)
    assert event.end == event.start + timedelta(days=1)
    assert event.duration == timedelta(days=1)


incomplete_event_1 = Event()
incomplete_event_2 = Event()
incomplete_event_2.add("DURATION", timedelta(hours=1))
incomplete_todo_1 = Todo()
incomplete_todo_2 = Todo()
incomplete_todo_2.add("DURATION", timedelta(hours=1))


@pytest.mark.parametrize(
    "incomplete_event_end",
    [
        incomplete_event_1,
        incomplete_todo_1,
    ],
)
@pytest.mark.parametrize("attr", ["start", "end", "duration"])
def test_incomplete_event(incomplete_event_end, attr):
    """Test that components without required properties throw the right error."""
    with pytest.raises(IncompleteComponent):
        getattr(incomplete_event_end, attr)


@pytest.mark.parametrize(
    "component_with_duration",
    [
        incomplete_event_2,  # Event with DURATION property
        incomplete_todo_2,  # Todo with DURATION property
    ],
)
def test_duration_property_accessible_without_dtstart(component_with_duration):
    """Test that DURATION property is accessible even without DTSTART (fixes issue #867)."""
    # DURATION property should be accessible directly
    assert component_with_duration.duration == timedelta(hours=1)

    # But start and end should still raise errors for incomplete components
    with pytest.raises(IncompleteComponent):
        _ = component_with_duration.start
    with pytest.raises(IncompleteComponent):
        _ = component_with_duration.end


@pytest.mark.parametrize(
    "invalid_value",
    [
        object(),
        timedelta(days=1),
        (datetime(2024, 10, 11, 10, 20), timedelta(days=1)),
    ],
)
@pytest.mark.parametrize(
    ("Component", "attr"),
    [
        (Event, "start"),
        (Event, "end"),
        (Event, "DTSTART"),
        (Event, "DTEND"),
        (Journal, "start"),
        (Journal, "end"),
        (Journal, "DTSTART"),
        (Todo, "start"),
        (Todo, "end"),
        (Todo, "DTSTART"),
        (Todo, "DUE"),
    ],
)
def test_set_invalid_start(invalid_value, attr, Component):
    """Check that we get the right error.

    - other types that vDDDTypes accepts
    - object
    """
    component = Component()
    with pytest.raises(TypeError) as e:
        setattr(component, attr, invalid_value)
    assert (
        e.value.args[0] == f"Use datetime or date, not {type(invalid_value).__name__}."
    )


def setitem(d: dict, key, value):
    d[key] = value


@pytest.mark.parametrize(
    "invalid_value",
    [
        object(),
        None,
        (datetime(2024, 10, 11, 10, 20), timedelta(days=1)),
        date(2012, 2, 2),
        datetime(2022, 2, 2),
    ],
)
def test_check_invalid_duration(start_end_component, invalid_value):
    """Check that we get the right error."""
    start_end_component["DURATION"] = invalid_value
    with pytest.raises(InvalidCalendar) as e:
        start_end_component.DURATION  # noqa: B018
    assert (
        e.value.args[0]
        == f"DURATION must be a timedelta, not {type(invalid_value).__name__}."
    )


def test_setting_the_end_deletes_the_duration(start_end_component):
    """Setting the end should not break the event."""
    DTEND = prop(start_end_component, "DTEND")
    start_end_component.DTSTART = datetime(2024, 10, 11, 10, 20)
    start_end_component.DURATION = timedelta(days=1)
    setattr(start_end_component, DTEND, datetime(2024, 10, 11, 10, 21))
    assert "DURATION" not in start_end_component
    assert start_end_component.DURATION is None
    end = getattr(start_end_component, DTEND)
    assert end == datetime(2024, 10, 11, 10, 21)


def test_setting_duration_deletes_the_end(start_end_component):
    """Setting the duration should not break the event."""
    DTEND = prop(start_end_component, "DTEND")
    start_end_component.DTSTART = datetime(2024, 10, 11, 10, 20)
    setattr(start_end_component, DTEND, datetime(2024, 10, 11, 10, 21))
    start_end_component.DURATION = timedelta(days=1)
    assert DTEND not in start_end_component
    assert getattr(start_end_component, DTEND) is None
    assert timedelta(days=1) == start_end_component.DURATION


valid_values = pytest.mark.parametrize(
    ("attr", "value"),
    [
        ("DTSTART", datetime(2024, 10, 11, 10, 20)),
        ("DTEND", datetime(2024, 10, 11, 10, 20)),
        ("DURATION", timedelta(days=1)),
    ],
)


@valid_values
def test_setting_to_none_deletes_value(start_end_component, attr, value):
    """Setting attributes to None deletes them."""
    attr = prop(start_end_component, attr)
    setattr(start_end_component, attr, value)
    assert attr in start_end_component
    assert getattr(start_end_component, attr) == value
    setattr(start_end_component, attr, None)
    assert attr not in start_end_component


@valid_values
def test_setting_a_value_twice(start_end_component, attr, value):
    """Setting attributes twice replaces them."""
    attr = prop(start_end_component, attr)
    setattr(start_end_component, attr, value + timedelta(days=1))
    setattr(start_end_component, attr, value)
    assert getattr(start_end_component, attr) == value


@pytest.mark.parametrize("attr", ["DTSTART", "DTEND", "DURATION"])
def test_invalid_none(start_end_component, attr):
    """Special case for None."""
    attr = prop(start_end_component, attr)
    start_end_component[attr] = None
    with pytest.raises(InvalidCalendar):
        getattr(start_end_component, attr)


def test_delete_duration(start_end_component):
    """Test the del command."""
    start_end_component.DURATION = timedelta(days=1)
    del start_end_component.DURATION
    assert start_end_component.DURATION is None


@pytest.mark.parametrize("attr", ["DTSTART", "end", "start"])
@pytest.mark.parametrize(
    "start",
    [
        datetime(2024, 10, 11, 10, 20),
        date(2024, 10, 11),
        datetime(2024, 10, 11, 10, 20, tzinfo=ZoneInfo("Europe/Paris")),
    ],
)
def test_journal_start(start, attr):
    """Test that we can set the start of a journal."""
    j = Journal()
    setattr(j, attr, start)
    assert start == j.DTSTART
    assert j.start == start
    assert j.end == start
    assert j.duration == timedelta(0)


@pytest.mark.parametrize("attr", ["start", "end"])
def test_delete_journal_start(attr):
    """Delete the start of the journal."""
    j = Journal()
    j.start = datetime(2010, 11, 12, 13, 14)
    j.DTSTART = None
    assert j.DTSTART is None
    assert "DTSTART" not in j
    with pytest.raises(IncompleteComponent):
        getattr(j, attr)


def setting_twice_does_not_duplicate_the_entry():
    j = Journal()
    j.DTSTART = date(2024, 1, 1)
    j.DTSTART = date(2024, 1, 3)
    assert date(2024, 1, 3) == j.DTSTART
    assert j.start == date(2024, 1, 3)
    assert j.end == date(2024, 1, 3)


@pytest.mark.parametrize(
    ("file", "trigger", "related"),
    [
        (
            "rfc_5545_absolute_alarm_example",
            vDatetime.from_ical("19970317T133000Z"),
            "START",
        ),
        ("rfc_5545_end", timedelta(days=-2), "END"),
        ("start_date", timedelta(days=-2), "START"),
    ],
)
def test_get_alarm_trigger_property(alarms, file, trigger, related):
    """Get the trigger property."""
    alarm = alarms[file]
    assert trigger == alarm.TRIGGER
    assert related == alarm.TRIGGER_RELATED


def test_set_alarm_trigger():
    """Set the alarm trigger."""
    a = Alarm()
    a.TRIGGER = timedelta(hours=1)
    assert timedelta(hours=1) == a.TRIGGER
    assert a.TRIGGER_RELATED == "START"


def test_set_alarm_trigger_related():
    """Set the alarm trigger."""
    a = Alarm()
    a.TRIGGER = timedelta(hours=1)
    a.TRIGGER_RELATED = "END"
    assert timedelta(hours=1) == a.TRIGGER
    assert a.TRIGGER_RELATED == "END"


def test_get_related_without_trigger():
    """The default is start"""
    assert Alarm().TRIGGER_RELATED == "START"


def test_cannot_set_related_without_trigger():
    """TRIGGER must be set to set the parameter."""
    with pytest.raises(ValueError) as e:
        a = Alarm()
        a.TRIGGER_RELATED = "END"
    assert (
        e.value.args[0]
        == "You must set a TRIGGER before setting the RELATED parameter."
    )


@pytest.mark.parametrize(
    ("file", "triggers"),
    [
        (
            "rfc_5545_absolute_alarm_example",
            (
                (),
                (),
                (
                    vDatetime.from_ical("19970317T133000Z"),
                    vDatetime.from_ical("19970317T134500Z"),
                    vDatetime.from_ical("19970317T140000Z"),
                    vDatetime.from_ical("19970317T141500Z"),
                    vDatetime.from_ical("19970317T143000Z"),
                ),
            ),
        ),
        ("rfc_5545_end", ((), (timedelta(days=-2),), ())),
        ("start_date", ((timedelta(days=-2),), (), ())),
    ],
)
def test_get_alarm_triggers(alarms, file, triggers):
    """Get the trigger property."""
    alarm = alarms[file]
    print(tuple(alarm.triggers))
    print(triggers)
    assert alarm.triggers == triggers


def test_triggers_emtpy_alarm():
    """An alarm with no trigger has no triggers."""
    assert Alarm().triggers == ((), (), ())


h1 = timedelta(hours=1)


def test_triggers_emtpy_with_no_repeat():
    """Check incomplete values."""
    a = Alarm()
    a.TRIGGER = h1
    a.DURATION = h1
    assert a.triggers == ((h1,), (), ())


def test_triggers_emtpy_with_no_duration():
    """Check incomplete values."""
    a = Alarm()
    a.TRIGGER = h1
    a.REPEAT = 10
    assert a.triggers == ((h1,), (), ())


@pytest.mark.parametrize(
    ("file", "triggers"),
    [
        (
            "rfc_5545_absolute_alarm_example",
            ((), (), (vDatetime.from_ical("19970317T133000Z"),)),
        ),
        ("rfc_5545_end", ((), (timedelta(days=-2),), ())),
        ("start_date", ((timedelta(days=-2),), (), ())),
    ],
)
@pytest.mark.parametrize("duration", [timedelta(days=-1), h1])
@pytest.mark.parametrize("repeat", [1, 3])
def test_get_alarm_triggers_repeated(alarms, file, triggers, duration, repeat):
    """Get the trigger property."""
    alarm = alarms[file].copy()
    alarm.REPEAT = repeat
    alarm.DURATION = duration
    for expected, triggers in zip(triggers, alarm.triggers):
        if not expected:
            assert triggers == ()
            continue
        assert len(triggers) == 1 + repeat
        assert triggers[0] == expected[0]
        for x, y in zip(triggers[:-1], triggers[1:]):
            assert y - x == duration



import datetime

import dateutil.parser

import icalendar
import icalendar.cal.calendar
import icalendar.cal.event
import icalendar.cal.timezone
from icalendar.prop import tzid_from_dt


def test_create_from_ical(calendars, other_tzp):
    """Create a calendar from a .ics file."""
    cal = calendars.timezoned

    assert cal["prodid"].to_ical() == b"-//Plone.org//NONSGML plone.app.event//EN"

    timezones = cal.walk("VTIMEZONE")
    assert len(timezones) == 1

    tz = timezones[0]
    assert tz["tzid"].to_ical() == b"Europe/Vienna"

    std = tz.walk("STANDARD")[0]
    assert std.decoded("TZOFFSETFROM") == datetime.timedelta(0, 7200)

    ev1 = cal.walk("VEVENT")[0]
    assert ev1.decoded("DTSTART") == other_tzp.localize(
        datetime.datetime(2012, 2, 13, 10, 0, 0), "Europe/Vienna"
    )
    assert ev1.decoded("DTSTAMP") == other_tzp.localize(
        datetime.datetime(2010, 10, 10, 9, 10, 10), "UTC"
    )


def test_create_to_ical(tzp):
    cal = icalendar.cal.calendar.Calendar()

    cal.add("prodid", "-//Plone.org//NONSGML plone.app.event//EN")
    cal.add("version", "2.0")
    cal.add("x-wr-calname", "test create calendar")
    cal.add("x-wr-caldesc", "icalendar tests")
    cal.add("x-wr-relcalid", "12345")
    cal.add("x-wr-timezone", "Europe/Vienna")

    tzc = icalendar.cal.timezone.Timezone()
    tzc.add("tzid", "Europe/Vienna")
    tzc.add("x-lic-location", "Europe/Vienna")

    tzs = icalendar.cal.timezone.TimezoneStandard()
    tzs.add("tzname", "CET")
    tzs.add("dtstart", datetime.datetime(1970, 10, 25, 3, 0, 0))
    tzs.add("rrule", {"freq": "yearly", "bymonth": 10, "byday": "-1su"})
    tzs.add("TZOFFSETFROM", datetime.timedelta(hours=2))
    tzs.add("TZOFFSETTO", datetime.timedelta(hours=1))

    tzd = icalendar.cal.timezone.TimezoneDaylight()
    tzd.add("tzname", "CEST")
    tzd.add("dtstart", datetime.datetime(1970, 3, 29, 2, 0, 0))
    tzs.add("rrule", {"freq": "yearly", "bymonth": 3, "byday": "-1su"})
    tzd.add("TZOFFSETFROM", datetime.timedelta(hours=1))
    tzd.add("TZOFFSETTO", datetime.timedelta(hours=2))

    tzc.add_component(tzs)
    tzc.add_component(tzd)
    cal.add_component(tzc)

    event = icalendar.cal.event.Event()
    event.add(
        "dtstart",
        tzp.localize(datetime.datetime(2012, 2, 13, 10, 00, 00), "Europe/Vienna"),
    )
    event.add(
        "dtend",
        tzp.localize(datetime.datetime(2012, 2, 17, 18, 00, 00), "Europe/Vienna"),
    )
    event.add(
        "dtstamp",
        tzp.localize(datetime.datetime(2010, 10, 10, 10, 10, 10), "Europe/Vienna"),
    )
    event.add(
        "created",
        tzp.localize(datetime.datetime(2010, 10, 10, 10, 10, 10), "Europe/Vienna"),
    )
    event.add("uid", "123456")
    event.add(
        "last-modified",
        tzp.localize(datetime.datetime(2010, 10, 10, 10, 10, 10), "Europe/Vienna"),
    )
    event.add("summary", "artsprint 2012")
    # event.add('rrule', 'FREQ=YEARLY;INTERVAL=1;COUNT=10')
    event.add("description", "sprinting at the artsprint")
    event.add("location", "aka bild, wien")
    event.add("categories", "first subject")
    event.add("categories", "second subject")
    event.add("attendee", "häns")
    event.add("attendee", "franz")
    event.add("attendee", "sepp")
    event.add("contact", "Max Mustermann, 1010 Wien")
    event.add("url", "https://plone.org")
    cal.add_component(event)

    test_out = b"|".join(cal.to_ical().splitlines())
    test_out = test_out.decode("utf-8")

    vtimezone_lines = "BEGIN:VTIMEZONE|TZID:Europe/Vienna|X-LIC-LOCATION:"
    "Europe/Vienna|BEGIN:STANDARD|DTSTART:19701025T03"
    "0000|RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=10|RRULE:FREQ=YEARLY;B"
    "YDAY=-1SU;BYMONTH=3|TZNAME:CET|TZOFFSETFROM:+0200|TZOFFSETTO:+01"
    "00|END:STANDARD|BEGIN:DAYLIGHT|DTSTART:19700329T"
    "020000|TZNAME:CEST|TZOFFSETFROM:+0100|TZOFFSETTO:+0200|END:DAYLI"
    "GHT|END:VTIMEZONE"
    assert vtimezone_lines in test_out

    test_str = "DTSTART;TZID=Europe/Vienna:20120213T100000"
    assert test_str in test_out
    assert "ATTENDEE:sepp" in test_out

    # ical standard expects DTSTAMP and CREATED in UTC
    assert "DTSTAMP:20101010T081010Z" in test_out
    assert "CREATED:20101010T081010Z" in test_out


def test_tzinfo_dateutil():
    """Test for issues #77, #63
    references: #73,7430b66862346fe3a6a100ab25e35a8711446717
    """
    date = dateutil.parser.parse("2012-08-30T22:41:00Z")
    date2 = dateutil.parser.parse("2012-08-30T22:41:00 +02:00")
    assert date.tzinfo.__module__.startswith("dateutil.tz")
    assert date2.tzinfo.__module__.startswith("dateutil.tz")

    # make sure, it's parsed properly and doesn't throw an error
    assert icalendar.vDDDTypes(date).to_ical() == b"20120830T224100Z"
    assert icalendar.vDDDTypes(date2).to_ical() == b"20120830T224100"


def test_create_america_new_york(calendars, tzp):
    """testing America/New_York, the most complex example from the RFC"""
    cal = calendars.america_new_york
    dt = cal.events[0].start
    assert tzid_from_dt(dt) in ("custom_America/New_York", "EDT")


def test_create_america_new_york_forward_reference(calendars, tzp):
    """testing America/New_York variant with VTIMEZONE as a forward reference"""
    cal = calendars.america_new_york_forward_reference
    dt = cal.walk("VEVENT")[0]["DTSTART"][0].dt
    assert tzid_from_dt(dt) in ("custom_America/New_York_Forward_reference", "EDT")


def test_america_new_york_with_pytz(calendars, tzp, pytz_only):
    """Create a custom timezone with pytz and test the transition times."""
    print(tzp)
    cal = calendars.america_new_york
    dt = cal.events[0].start
    tz = dt.tzinfo
    tz_new_york = tzp.timezone("America/New_York")
    # for reasons (tm) the locally installed version of the timezone
    # database isn't always complete, therefore we only compare some
    # transition times
    ny_transition_times = []
    ny_transition_info = []
    for num, date in enumerate(tz_new_york._utc_transition_times):
        if (
            datetime.datetime(1967, 4, 30, 7, 0)
            <= date
            <= datetime.datetime(2037, 11, 1, 6, 0)
        ):
            ny_transition_times.append(date)
            ny_transition_info.append(tz_new_york._transition_info[num])
    assert tz._utc_transition_times[:142] == ny_transition_times
    assert tz._transition_info[0:142] == ny_transition_info
    assert (
        datetime.timedelta(-1, 72000),
        datetime.timedelta(0, 3600),
        "EDT",
    ) in tz._tzinfos
    assert (
        datetime.timedelta(-1, 68400),
        datetime.timedelta(0),
        "EST",
    ) in tz._tzinfos


fiji_transition_times = [
    datetime.datetime(1915, 10, 25, 12, 4),
    datetime.datetime(1998, 10, 31, 14, 0),
    datetime.datetime(1999, 2, 27, 14, 0),
    datetime.datetime(1999, 11, 6, 14, 0),
    datetime.datetime(2000, 2, 26, 14, 0),
    datetime.datetime(2009, 11, 28, 14, 0),
    datetime.datetime(2010, 3, 27, 14, 0),
    datetime.datetime(2010, 10, 23, 14, 0),
    datetime.datetime(2011, 3, 5, 14, 0),
    datetime.datetime(2011, 10, 22, 14, 0),
    datetime.datetime(2012, 1, 21, 14, 0),
    datetime.datetime(2012, 10, 20, 14, 0),
    datetime.datetime(2013, 1, 19, 14, 0),
    datetime.datetime(2013, 10, 26, 14, 0),
    datetime.datetime(2014, 1, 18, 13, 0),
    datetime.datetime(2014, 10, 25, 14, 0),
    datetime.datetime(2015, 1, 17, 13, 0),
    datetime.datetime(2015, 10, 24, 14, 0),
    datetime.datetime(2016, 1, 23, 13, 0),
    datetime.datetime(2016, 10, 22, 14, 0),
    datetime.datetime(2017, 1, 21, 13, 0),
    datetime.datetime(2017, 10, 21, 14, 0),
    datetime.datetime(2018, 1, 20, 13, 0),
    datetime.datetime(2018, 10, 20, 14, 0),
    datetime.datetime(2019, 1, 19, 13, 0),
    datetime.datetime(2019, 10, 26, 14, 0),
    datetime.datetime(2020, 1, 18, 13, 0),
    datetime.datetime(2020, 10, 24, 14, 0),
    datetime.datetime(2021, 1, 23, 13, 0),
    datetime.datetime(2021, 10, 23, 14, 0),
    datetime.datetime(2022, 1, 22, 13, 0),
    datetime.datetime(2022, 10, 22, 14, 0),
    datetime.datetime(2023, 1, 21, 13, 0),
    datetime.datetime(2023, 10, 21, 14, 0),
    datetime.datetime(2024, 1, 20, 13, 0),
    datetime.datetime(2024, 10, 26, 14, 0),
    datetime.datetime(2025, 1, 18, 13, 0),
    datetime.datetime(2025, 10, 25, 14, 0),
    datetime.datetime(2026, 1, 17, 13, 0),
    datetime.datetime(2026, 10, 24, 14, 0),
    datetime.datetime(2027, 1, 23, 13, 0),
    datetime.datetime(2027, 10, 23, 14, 0),
    datetime.datetime(2028, 1, 22, 13, 0),
    datetime.datetime(2028, 10, 21, 14, 0),
    datetime.datetime(2029, 1, 20, 13, 0),
    datetime.datetime(2029, 10, 20, 14, 0),
    datetime.datetime(2030, 1, 19, 13, 0),
    datetime.datetime(2030, 10, 26, 14, 0),
    datetime.datetime(2031, 1, 18, 13, 0),
    datetime.datetime(2031, 10, 25, 14, 0),
    datetime.datetime(2032, 1, 17, 13, 0),
    datetime.datetime(2032, 10, 23, 14, 0),
    datetime.datetime(2033, 1, 22, 13, 0),
    datetime.datetime(2033, 10, 22, 14, 0),
    datetime.datetime(2034, 1, 21, 13, 0),
    datetime.datetime(2034, 10, 21, 14, 0),
    datetime.datetime(2035, 1, 20, 13, 0),
    datetime.datetime(2035, 10, 20, 14, 0),
    datetime.datetime(2036, 1, 19, 13, 0),
    datetime.datetime(2036, 10, 25, 14, 0),
    datetime.datetime(2037, 1, 17, 13, 0),
    datetime.datetime(2037, 10, 24, 14, 0),
    datetime.datetime(2038, 1, 23, 13, 0),
    datetime.datetime(2038, 10, 23, 14, 0),
]

fiji_transition_info = (
    [
        (
            datetime.timedelta(0, 43200),
            datetime.timedelta(0),
            "custom_Pacific/Fiji_19151026T000000_+115544_+1200",
        )
    ]
    + 3
    * [
        (
            datetime.timedelta(0, 46800),
            datetime.timedelta(0, 3600),
            "custom_Pacific/Fiji_19981101T020000_+1200_+1300",
        ),
        (
            datetime.timedelta(0, 43200),
            datetime.timedelta(0),
            "custom_Pacific/Fiji_19990228T030000_+1300_+1200",
        ),
    ]
    + 3
    * [
        (
            datetime.timedelta(0, 46800),
            datetime.timedelta(0, 3600),
            "custom_Pacific/Fiji_20101024T020000_+1200_+1300",
        ),
        (
            datetime.timedelta(0, 43200),
            datetime.timedelta(0),
            "custom_Pacific/Fiji_19990228T030000_+1300_+1200",
        ),
    ]
    + 25
    * [
        (
            datetime.timedelta(0, 46800),
            datetime.timedelta(0, 3600),
            "custom_Pacific/Fiji_20101024T020000_+1200_+1300",
        ),
        (
            datetime.timedelta(0, 43200),
            datetime.timedelta(0),
            "custom_Pacific/Fiji_20140119T020000_+1300_+1200",
        ),
    ]
    + [
        (
            datetime.timedelta(0, 46800),
            datetime.timedelta(0, 3600),
            "custom_Pacific/Fiji_20101024T020000_+1200_+1300",
        )
    ]
)


def test_create_pacific_fiji(calendars, pytz_only):
    """testing Pacific/Fiji, another pretty complex example with more than
    one RDATE property per subcomponent"""
    cal = calendars.pacific_fiji

    tz = cal.walk("VEVENT")[0]["DTSTART"][0].dt.tzinfo
    assert str(tz) == "custom_Pacific/Fiji"
    assert tz._utc_transition_times == fiji_transition_times
    assert tz._transition_info == fiji_transition_info
    assert (
        datetime.timedelta(0, 46800),
        datetime.timedelta(0, 3600),
        "custom_Pacific/Fiji_19981101T020000_+1200_+1300",
    ) in tz._tzinfos
    assert (
        datetime.timedelta(0, 43200),
        datetime.timedelta(0),
        "custom_Pacific/Fiji_19990228T030000_+1300_+1200",
    ) in tz._tzinfos


# these are the expected offsets before and after the fiji_transition_times
fiji_expected_offsets = [
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
    (datetime.timedelta(hours=13), datetime.timedelta(hours=13)),
    (datetime.timedelta(hours=12), datetime.timedelta(hours=12)),
]


def test_transition_times_fiji(tzp, timezones):
    """The transition times are computed."""
    tz = timezones.pacific_fiji.to_tz(tzp, lookup_tzid=False)
    offsets = []  # [(before, after), ...]
    for _i, transition_time in enumerate(fiji_transition_times):
        before_after_offset = []
        for offset in (datetime.timedelta(hours=-1), datetime.timedelta(hours=+1)):
            time_in_timezone = tzp.localize(transition_time + offset, tz)
            utc_offset = time_in_timezone.utcoffset()
            before_after_offset.append(utc_offset)
        offsets.append(tuple(before_after_offset))
    assert offsets == fiji_expected_offsets


def test_same_start_date(calendars):
    """testing if we can handle VTIMEZONEs whose different components
    have the same start DTIMEs."""
    cal = calendars.timezone_same_start
    d = cal.subcomponents[1]["DTSTART"].dt
    assert d.strftime("%c") == "Fri Feb 24 12:00:00 2017"


def test_same_start_date_and_offset(calendars):
    """testing if we can handle VTIMEZONEs whose different components
    have the same DTSTARTs, TZOFFSETFROM, and TZOFFSETTO."""
    cal = calendars.timezone_same_start_and_offset
    d = cal.subcomponents[1]["DTSTART"].dt
    assert d.strftime("%c") == "Fri Feb 24 12:00:00 2017"


def test_rdate(calendars):
    """testing if we can handle VTIMEZONEs who only have an RDATE, not RRULE"""
    cal = calendars.timezone_rdate
    vevent = cal.walk("VEVENT")[0]
    assert tzid_from_dt(vevent["DTSTART"].dt) in ("posix/Europe/Vaduz", "CET")


def test_rdate_pytz(calendars, pytz_only):
    """testing if we can handle VTIMEZONEs who only have an RDATE, not RRULE"""
    cal = calendars.timezone_rdate
    vevent = cal.walk("VEVENT")[0]
    tz = vevent["DTSTART"].dt.tzinfo
    assert tz._utc_transition_times[:6] == [
        datetime.datetime(1901, 12, 13, 20, 45, 38),
        datetime.datetime(1941, 5, 5, 0, 0, 0),
        datetime.datetime(1941, 10, 6, 0, 0, 0),
        datetime.datetime(1942, 5, 4, 0, 0, 0),
        datetime.datetime(1942, 10, 5, 0, 0, 0),
        datetime.datetime(1981, 3, 29, 1, 0),
    ]
    assert tz._transition_info[:6] == [
        (datetime.timedelta(0, 3600), datetime.timedelta(0), "CET"),
        (datetime.timedelta(0, 7200), datetime.timedelta(0, 3600), "CEST"),
        (datetime.timedelta(0, 3600), datetime.timedelta(0), "CET"),
        (datetime.timedelta(0, 7200), datetime.timedelta(0, 3600), "CEST"),
        (datetime.timedelta(0, 3600), datetime.timedelta(0), "CET"),
        (datetime.timedelta(0, 7200), datetime.timedelta(0, 3600), "CEST"),
    ]



"""Test the image class to convert from and to binary data."""

from __future__ import annotations

import base64

import pytest

from icalendar import Calendar, Image, vBinary, vUri
from icalendar.prop import vText

TRANSPARENT_PIXEL = base64.b64decode("""iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCA
YAAAAfFcSJAAAACXBIWXMAAAAnAAAAJwEqCZFPAAAAGXRFWHRTb2Z0d2FyZQB3d3cuaW5rc2NhcGUub3Jn
m+48GgAAAA1JREFUCJlj+P//PwMACPwC/oXNqzQAAAAASUVORK5CYII=""")


@pytest.fixture
def images(calendars) -> dict[str, Image]:
    """Return the images we get from the example calendars."""
    calendar: Calendar = calendars.rfc_7986_image
    images: dict[str, Image] = {}
    for component in calendar.subcomponents:
        img = component.images[0]
        images[component.uid] = img
    return images


@pytest.mark.parametrize(
    ("uid", "uri", "data", "fmt_type", "altrep", "display"),
    [
        (
            "uri-event",
            "http://example.com/images/party.png",
            None,
            "image/png",
            None,
            "BADGE",
        ),
        (
            "uri-todo",
            "http://example.com/images/party.jpg",
            None,
            None,
            None,
            "BADGE",
        ),
        (
            "uri-journal",
            "http://example.com/images/party.jpg",
            None,
            "image/jpg",
            None,
            "ICON",
        ),
        (
            "data-event",
            None,
            TRANSPARENT_PIXEL,
            "image/png",
            None,
            None,
        ),
        (
            "data-todo",
            None,
            TRANSPARENT_PIXEL,
            "image/png",
            "http://example.com/images/party.jpg",
            None,
        ),
        (
            "data-journal",
            None,
            TRANSPARENT_PIXEL,
            "image/png",
            None,
            "BADGE",
        ),
    ],
)
def test_image_parsing(
    images,
    uid,
    uri,
    data,
    fmt_type,
    altrep,
    display,
):
    """Test that the image property is parsed correctly."""
    image = images[uid]
    assert image.uri == uri
    assert image.data == data
    assert image.fmttype == fmt_type
    assert image.altrep == altrep
    assert image.display == display


def test_no_images():
    """Test that an empty calendar has no images."""
    calendar = Calendar()
    assert len(calendar.images) == 0


def test_create_image_invalid_type():
    """Test that creating an image with invalid type raises TypeError."""
    with pytest.raises(TypeError):
        Image.from_property_value("not a valid type")
    with pytest.raises(TypeError):
        Image.from_property_value(vText("not a valid type", params={"VALUE": "TEXT"}))
    with pytest.raises(TypeError):
        Image.from_property_value(vText("http://example.com/image.png"))


def test_create_with_vBinary():
    """Test creating an Image from a vBinary property."""
    b64data = base64.b64encode(TRANSPARENT_PIXEL).decode("ascii")
    vbin = vBinary(b64data, params={"FMTTYPE": "image/png"})
    image = Image.from_property_value(vbin)
    assert image.uri is None
    assert image.data == TRANSPARENT_PIXEL
    assert image.fmttype == "image/png"
    assert image.altrep is None
    assert image.display is None


def test_create_with_vUri():
    """Test creating an Image from a vUri property."""
    uri = "http://example.com/image.png"
    vuri = vUri(uri, params={"FMTTYPE": "image/png", "DISPLAY": "BADGE"})
    image = Image.from_property_value(vuri)
    assert image.uri == uri
    assert image.data is None
    assert image.fmttype == "image/png"
    assert image.altrep is None
    assert image.display == "BADGE"


def test_create_image_with_vText_as_uri():
    """Test that creating an image with vText but VALUE URI or BINARY raises TypeError."""
    img = Image.from_property_value(
        vText("http://example.com/image.png", params={"VALUE": "URI"})
    )
    assert img.uri == "http://example.com/image.png"
    assert img.data is None
    assert img.fmttype is None
    assert img.altrep is None
    assert img.display is None


def test_create_image_with_vText_as_binary():
    """Test that creating an image with vText but VALUE URI or BINARY raises TypeError."""
    b64data = base64.b64encode(TRANSPARENT_PIXEL).decode("ascii")
    img = Image.from_property_value(vText(b64data, params={"VALUE": "BINARY"}))
    assert img.uri is None
    assert img.data == TRANSPARENT_PIXEL
    assert img.fmttype is None
    assert img.altrep is None
    assert img.display is None


def test_requires_uri_xor_binary():
    """Test forbidden parameter combinations."""
    with pytest.raises(ValueError):
        Image()
    with pytest.raises(ValueError):
        Image(b64data="", uri="http://example.com/image.png")



"""This tests the exdate property."""

from __future__ import annotations

from datetime import date, datetime
from typing import Union

import pytest

from icalendar import (
    Event,
    Journal,
    TimezoneDaylight,
    TimezoneStandard,
    Todo,
)

C_EXDATE = Union[Event, Todo, Journal, TimezoneDaylight, TimezoneStandard]


@pytest.fixture(params=[Event, Todo, Journal, TimezoneDaylight, TimezoneStandard])
def c_exdate(request) -> C_EXDATE:
    """Return a component that uses exdate."""
    return request.param()


@pytest.fixture(
    params=[
        lambda _tzp: date(2019, 10, 11),
        lambda _tzp: datetime(2000, 1, 13, 12, 1),
        lambda tzp: tzp.localize_utc(datetime(2031, 12, 1, 23, 59)),
        lambda tzp: tzp.localize(datetime(1984, 1, 13, 13, 1), "Europe/Athens"),
    ]
)
def exdate(request, tzp):
    """Possible values for an exdate."""
    return request.param(tzp)


def test_no_exdates_by_default(c_exdate):
    """We expect no exdate by default."""
    assert c_exdate.exdates == []


def test_set_and_retrieve_exdate(exdate, c_exdate):
    """Set the attribute and get the value."""
    c_exdate.add("exdate", [exdate])
    result = [exdate]
    assert c_exdate.exdates == result


def test_set_and_retrieve_exdates_in_list(exdate, c_exdate):
    """Set the attribute and get the value."""
    c_exdate.add("exdate", [exdate, exdate])
    result = [exdate, exdate]
    assert c_exdate.exdates == result


def test_set_and_retrieve_exdates_twice(exdate, c_exdate):
    """Set the attribute and get the value."""
    c_exdate.add("exdate", [exdate])
    c_exdate.add("exdate", [exdate])
    result = [exdate, exdate]
    assert c_exdate.exdates == result



"""Access to examples of components."""

from __future__ import annotations

from pathlib import Path


def get_example(component_directory: str, example_name: str) -> bytes:
    """Return an example and raise an error if it is absent."""
    here = Path(__file__).parent.parent
    examples = here / "tests" / component_directory
    if not example_name.endswith(".ics"):
        example_name = example_name + ".ics"
    example_file = examples / example_name
    if not example_file.is_file():
        raise ValueError(
            f"Example {example_name} for {component_directory} not found. "
            f"You can use one of {', '.join(p.name for p in examples.iterdir())}"
        )
    return Path(example_file).read_bytes()


__all__ = ["get_example"]



"""Parse the calendar and make sure the timezone is used.

See https://github.com/collective/icalendar/issues/218
"""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import pytest

from icalendar.timezone.tzid import tzid_from_dt

if TYPE_CHECKING:
    from icalendar.cal.event import Event


@pytest.fixture
def event(calendars) -> Event:
    """The event to check."""
    return calendars.issue_218_bad_tzid.events[0]


def test_event_has_start_and_end(event: Event):
    """The calendar should be parsed and the start and end have a timezone."""
    assert event.start.replace(tzinfo=None) == datetime(2017, 2, 28, 23, 00)
    assert event.end.replace(tzinfo=None) == datetime(2017, 2, 28, 23, 30)


def test_timezone(event: Event):
    """The event uses a timezone."""
    assert event.start.tzinfo == event.end.tzinfo
    assert tzid_from_dt(event.start) == "UTC+11"
    assert event.start.utcoffset() == timedelta(hours=11)



"""Test for issue #898 - TODO.end should equal TODO.start when no DURATION/DUE.

For Events with only DTSTART:
- date: end = start + 1 day
- datetime: end = start

For Todos with only DTSTART:
- date: end = start (NOT start + 1 day)
- datetime: end = start

See https://github.com/collective/icalendar/issues/898
"""

from datetime import date, datetime

from icalendar import Event, Todo


class TestIssue898TodoEnd:
    """Test TODO.end behavior when only DTSTART is provided."""

    def test_event_with_date_dtstart_only(self):
        """Event with date DTSTART should have end = start + 1 day."""
        event = Event()
        event.add("UID", "test-event-date")
        event.add("DTSTART", date(2026, 1, 1))

        assert event.end == date(2026, 1, 2)

    def test_event_with_datetime_dtstart_only(self):
        """Event with datetime DTSTART should have end = start."""
        event = Event()
        event.add("UID", "test-event-datetime")
        event.add("DTSTART", datetime(2026, 1, 1, 12, 0))

        assert event.end == datetime(2026, 1, 1, 12, 0)

    def test_todo_with_date_dtstart_only(self):
        """Todo with date DTSTART should have end = start (NOT start + 1 day).

        The RFC doesn't specify this behavior explicitly.
        Returning None is another option, but returning start works best.
        """
        todo = Todo()
        todo.add("UID", "test-todo-date")
        todo.add("DTSTART", date(2026, 1, 1))

        # This should be start, not start + 1 day
        assert todo.end == date(2026, 1, 1)

    def test_todo_with_datetime_dtstart_only(self):
        """Todo with datetime DTSTART should have end = start."""
        todo = Todo()
        todo.add("UID", "test-todo-datetime")
        todo.add("DTSTART", datetime(2026, 1, 1, 12, 0))

        assert todo.end == datetime(2026, 1, 1, 12, 0)



import string
import unittest

import hypothesis.strategies as st
from hypothesis import given, settings

from icalendar.parser import Contentline, Contentlines, Parameters


def printable_characters(**kw):
    return st.text(st.characters(blacklist_categories=("Cc", "Cs"), **kw))


key = st.text(string.ascii_letters + string.digits, min_size=1)
value = printable_characters(blacklist_characters='\\;:"')


class TestFuzzing(unittest.TestCase):
    @given(
        lines=st.lists(st.tuples(key, st.dictionaries(key, value), value), min_size=1)
    )
    @settings(max_examples=10**3)
    def test_main(self, lines):
        cl = Contentlines()
        for key, params, value in lines:
            try:
                params = Parameters(**params)
            except TypeError:
                # Happens when there is a random parameter 'self'...
                continue
            cl.append(Contentline.from_parts(key, params, value))
        cl.append("")

        assert Contentlines.from_ical(cl.to_ical()) == cl



"""Testing multiple VCALENDAR components and multiple VEVENT components"""

import pytest

from icalendar import Event, vText


def test_multiple(calendars):
    """Check opening multiple calendars."""

    cals = calendars.multiple.multiple_calendar_components

    assert len(cals) == 2
    assert [comp.name for comp in cals[0].walk()] == ["VCALENDAR", "VEVENT"]
    assert [comp.name for comp in cals[1].walk()] == ["VCALENDAR", "VEVENT", "VEVENT"]
    assert cals[0]["prodid"] == vText(
        "-//Mozilla.org/NONSGML Mozilla Calendar V1.0//EN"
    )


def test_multiple_events():
    """Raises ValueError unless multiple=True"""
    event_components = """
BEGIN:VEVENT
END:VEVENT
BEGIN:VEVENT
END:VEVENT
"""
    with pytest.raises(ValueError):
        Event.from_ical(event_components, multiple=False)


def test_missing_event():
    """Raises ValueError if no component found"""
    with pytest.raises(ValueError):
        Event.from_ical("")



"""This tests the UID property.

See https://github.com/collective/icalendar/issues/740
"""

from icalendar.cal.alarm import Alarm


def test_alarm_uses_x_property_too():
    alarm = Alarm()
    alarm["X-ALARMUID"] = "1234"
    assert alarm.uid == "1234"
